use druid::widget::prelude::*;
use druid::{Point, WidgetPod};

pub struct HiddenIf<T, W> {
    child: WidgetPod<T, W>,
    hidden_if: Box<dyn Fn(&T, &Env) -> bool>,
}

impl<T: Data, W: Widget<T>> HiddenIf<T, W> {
    pub fn new(child: W, hidden_if: impl Fn(&T, &Env) -> bool + 'static) -> Self {
        Self {
            hidden_if: Box::new(hidden_if),
            child: WidgetPod::new(child),
        }
    }
}

impl<T: Data, W: Widget<T>> Widget<T> for HiddenIf<T, W> {
    #[tracing::instrument(name = "HideContainer", level = "trace", skip_all)]
    fn event(&mut self, ctx: &mut EventCtx, event: &Event, data: &mut T, env: &Env) {
        let hidden = (self.hidden_if)(data, env);
        ctx.set_disabled(hidden);
        self.child.event(ctx, event, data, env);
    }

    #[tracing::instrument(name = "HideContainer", level = "trace", skip_all)]
    fn lifecycle(&mut self, ctx: &mut LifeCycleCtx, event: &LifeCycle, data: &T, env: &Env) {
        let hidden = (self.hidden_if)(data, env);
        ctx.set_disabled(hidden);
        self.child.lifecycle(ctx, event, data, env)
    }

    #[tracing::instrument(name = "HideContainer", level = "trace", skip_all)]
    fn update(&mut self, ctx: &mut UpdateCtx, _: &T, data: &T, env: &Env) {
        self.child.update(ctx, data, env);
    }

    #[tracing::instrument(name = "HideContainer", level = "trace", skip_all)]
    fn layout(&mut self, ctx: &mut LayoutCtx, bc: &BoxConstraints, data: &T, env: &Env) -> Size {
        bc.debug_check("HideContainer");
        let hidden = (self.hidden_if)(data, env);
        if hidden {
            return Size::ZERO;
        }

        let child_size = self.child.layout(ctx, bc, data, env);
        self.child.set_origin(ctx, Point::new(0.0, 0.0));
        child_size
    }

    #[tracing::instrument(name = "HideContainer", level = "trace", skip_all)]
    fn paint(&mut self, ctx: &mut PaintCtx, data: &T, env: &Env) {
        let hidden = (self.hidden_if)(data, env);
        if hidden {
            return;
        }

        self.child.paint(ctx, data, env);
    }
}
